(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 1997-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: generator.ml,v 1.19 2014/01/31 09:51:52 fclement Exp $ *)

open Og_types
;;

exception Found_recursive;;

let rec_flag_of_type_definition (ty_def_list : type_defs) =

  let defined_names =
    List.fold_left
      (fun accu (ty_name, _, _ : type_def) -> ty_name :: accu)
      [] ty_def_list in

   let find_rec_ty_ident ty_ident =
     match ty_ident with
     | Lident ty_name ->
       if List.mem ty_name defined_names then raise Found_recursive
     | Ldot (_, _) | Lapply (_, _) -> () in

  let find_rec_ty_in_type =
    let rec loop = function
      | Var_type _s -> ()
      | Const_type ty_ident -> find_rec_ty_ident ty_ident
      | Apply_type (ty_ident, fields) ->
        find_rec_ty_ident ty_ident;
        Array.iter loop fields
      | Tuple_type fields
      | Function_type fields -> Array.iter loop fields
      | Paren_type ty -> loop ty in
    loop in

  let find_rec_ty_in_record_field = function
    | Field (_label, ty)
    | Field_mutable (_label, ty) -> find_rec_ty_in_type ty
  and find_rec_ty_in_sum_field = function
    | Sum_tag _sum_label -> ()
    | Sum_inj (_sum_label, ty) -> find_rec_ty_in_type ty
  and find_rec_ty_in_alias_field = function
    | (ty, _kind_opt) -> find_rec_ty_in_type ty in

  let find_rec_ty_in_type_kind =
    let loop = function
      | Record_type (_vis, record_fields) ->
        List.iter find_rec_ty_in_record_field record_fields
      | Sum_type (_vis, sum_fields) ->
        List.iter find_rec_ty_in_sum_field sum_fields
      | Alias_type (_vis, alias_field) ->
        find_rec_ty_in_alias_field alias_field
      | External_type -> () in
    loop in

  let find_rec_ty_in_type_def (_ty_name, _ty_args, ty_kind) =
    find_rec_ty_in_type_kind ty_kind in

  try
    List.iter find_rec_ty_in_type_def ty_def_list;
    Walker_ast.No_rec
  with
  | Found_recursive ->
    Walker_ast.Rec
;;

(* Generating pattern variable names of a given type *)
let rec ty_name_prefix = function
  | Lident ty_name ->
    String.make 1 (Char.lowercase (Og_misc.string_first_char ty_name))
  | Ldot (longident, ty_name) ->
    let pref = ty_name_prefix longident in
    Printf.sprintf "%s_%c" pref (Og_misc.string_first_char ty_name)
  | Lapply (foncident, modident) ->
    Printf.sprintf "%s_%s"
      (ty_name_prefix foncident) (ty_name_prefix modident)
;;

let rec prefix_of_type ty =

  let rec loop accu = function
    | Var_type s -> String.make 1 (Og_misc.string_first_char s) :: accu
    | Const_type ty_name -> ty_name_prefix ty_name :: accu
    | Apply_type (ty_name, ty_v) ->
      let pref = ty_name_prefix ty_name in
      pref :: prefix_of_tuple_fields accu ty_v
    | Tuple_type fields -> prefix_of_tuple_fields ("t" :: accu) fields
    | Function_type fields ->
      prefix_of_function_fields ("f" :: accu) fields
    | Paren_type ty -> ("t_" ^ prefix_of_type ty) :: accu

  and prefix_of_tuple_fields accu fields =
    let rec loop accu i j =
      match compare j 1 with
      | 0 | 1 ->
        loop (prefix_of_type fields.(i) :: accu) (i + 1) (j - 1)
      | _ -> accu in
    loop accu 0 (Array.length fields)

  and prefix_of_function_fields accu fields =
    let rec loop accu i j =
      match compare j 1 with
      | 0 | 1 ->
        loop (prefix_of_type fields.(i) :: accu) (i + 1) (j - 1)
      | _ -> accu in
    loop accu 0 (Array.length fields) in

  String.concat "_" (List.rev (loop [] ty))
;;

let pattern_of_type ty =
  Walker_ast.Pvariable (Symbol.make (prefix_of_type ty), ty)
;;

let rec make_pattern ty =
  match ty with
  | Var_type _ -> pattern_of_type ty
  | Const_type _ -> pattern_of_type ty
  | Apply_type (_type_name, _type_expr_array) -> pattern_of_type ty
  | Tuple_type type_expr_array ->
    Walker_ast.Ptuple (Array.map (make_pattern) type_expr_array, ty)
  | Function_type _ -> pattern_of_type ty
  | Paren_type ty -> pattern_of_type ty
;;

let rec string_of_type_long_ident = function
  | Lident ty_name -> ty_name
  | Ldot (longident, ty_name) ->
    string_of_type_long_ident longident ^ "." ^ ty_name
  | Lapply (foncident, modident) ->
    string_of_type_long_ident foncident ^
    "(" ^ string_of_type_long_ident modident ^ ")"
;;

let find_ty_vars_in_ty_kind =
  let find_vars_in_ty accu =
    let rec loop accu = function
      | Var_type s -> if List.mem s accu then accu else s :: accu
      | Const_type _ty_name -> accu
      | Apply_type (_, fields)
      | Tuple_type fields
      | Function_type fields -> Array.fold_left loop accu fields
      | Paren_type ty -> loop accu ty in
    loop accu in

  let find_vars_in_record_field accu = function
    | Field (_label, ty)
    | Field_mutable (_label, ty) -> find_vars_in_ty accu ty
  and find_vars_in_sum_field accu = function
    | Sum_tag _sum_label -> accu
    | Sum_inj (_sum_label, ty) -> find_vars_in_ty accu ty
  and find_vars_in_alias_field accu = function
    | (ty, _kind_opt) -> find_vars_in_ty accu ty in

  let find_ty_vars accu = function
    | Record_type (_vis, record_fields) ->
      List.fold_left find_vars_in_record_field accu record_fields
    | Sum_type (_vis, sum_fields) ->
      List.fold_left find_vars_in_sum_field accu sum_fields
    | Alias_type (_vis, alias_field) ->
      find_vars_in_alias_field accu alias_field
    | External_type -> accu in

  fun ty_kind -> find_ty_vars [] ty_kind
;;

module Make (P: Generator_spec.Prefix) :
            Generator_spec.Walker_ast_generator =

struct


  (* Generating names for functions *)
  let default_walker_name s = P.function_prefix ^ s;;
  let local_walker_name = default_walker_name;;
  (* Generating names for functional arguments *)
  let walker_var_arg_name s = P.type_var_arg_prefix ^ s;;
  let unused_var_name s = "_" ^ s;;
  let walker_unused_var_arg_name s = unused_var_name (walker_var_arg_name s);;
  (* Generating names for external printer modules *)
  let external_walker_module_name s = s ^ P.external_module_suffix;;
  let default_inj_name inj_type = P.injection_prefix ^ inj_type;;
  let local_inj_name inj_type =
    Printf.sprintf "%s.%s"
      (Og_config.get_input_module_name ())
      (default_inj_name inj_type)
  ;;

  let external_inj_module_name s = s;;

  let buffer_arg = P.buffer_argument;;
  let buffer_argument =
    match buffer_arg with
    | None -> None
    | Some (arg, ty) -> Some (Walker_ast.Pvariable (arg, ty))
  ;;

  let bufferize_pattern p = buffer_argument, p;;

  let local_env = ref [];;

  let external_name default_name external_module_name = function
    | Lident ty_name -> default_name ty_name
    | Ldot (longident, ty_name) ->
      let mod_name =
        external_module_name (string_of_type_long_ident longident) in
      let name = default_name ty_name in
      Printf.sprintf "%s.%s" mod_name name
    | Lapply (foncident, modident) ->
      let fonc_name = string_of_type_long_ident foncident in
      let mod_name = string_of_type_long_ident modident in
      Printf.sprintf "%s (%s)" fonc_name mod_name
  ;;

(*
  let rec external_walker_name = function
    | Lident ty_name -> default_walker_name ty_name
    | Ldot (longident, ty_name) ->
      external_walker_module_name (string_of_type_long_ident longident) ^
      "." ^
      external_walker_name (Lident ty_name)
    | Lapply (foncident, modident) ->
      string_of_type_long_ident foncident ^
      "(" ^ string_of_type_long_ident modident ^ ")"
  ;;

  let rec external_inj_name = function
    | Lident ty_name -> inj_walker_name ty_name
    | Ldot (longident, ty_name) ->
      external_inj_module_name (string_of_type_long_ident longident) ^
      "." ^
      external_inj_name (Lident ty_name)
    | Lapply (foncident, modident) ->
      string_of_type_long_ident foncident ^
      "(" ^ string_of_type_long_ident modident ^ ")"
  ;;
*)
  let external_walker_name =
    external_name default_walker_name external_walker_module_name
  ;;

  let external_inj_name =
    external_name default_inj_name external_inj_module_name
  ;;

  let get_walker_name = function
    | Lident ty_name ->
      if List.mem_assoc ty_name !local_env
      then Walker_ast.Fname_local (List.assoc ty_name !local_env)
      else Walker_ast.Fname_external (default_walker_name ty_name)
    | Lapply (_, _) | Ldot (_, _) as lident ->
      Walker_ast.Fname_external (external_walker_name lident)
  ;;

  let get_inj_name = function
    | Lident ty_name ->
      if List.mem_assoc ty_name !local_env
      then Walker_ast.Fname_local (local_inj_name ty_name)
      else Walker_ast.Fname_external (default_inj_name ty_name)
    | Lapply (_, _) | Ldot (_, _) as lident ->
      Walker_ast.Fname_external (external_inj_name lident)
  ;;

  let rec walker_fun = function
    | Var_type s ->
      Walker_ast.Fconstant (Walker_ast.Fname_arg (walker_var_arg_name s))
    | Const_type lident ->
      Walker_ast.Fconstant (get_walker_name lident)
    | Apply_type (lident, type_expr_array) ->
      let p_name = get_walker_name lident in
      let par_walker_fun ty = paren_walker_fun (walker_fun ty) in
      let pats =
        Og_misc.array_map_to_list par_walker_fun type_expr_array in
      Walker_ast.Fapply (p_name, pats)
    | Tuple_type _ as ty -> make_walker_tuple_fun ty
    | Function_type type_expr_array ->
      let lident = Lident "fun" in
      walker_fun (Apply_type (lident, type_expr_array))
    | Paren_type ty ->
      paren_walker_fun (walker_fun ty)

  and paren_walker_fun fn =
    match fn with
    | Walker_ast.Fconstant _
    | Walker_ast.Fparen _ -> fn
    | Walker_ast.Fapply (_, _)
    | Walker_ast.Fabs (_, _) -> Walker_ast.Fparen fn

  and argument_of_pat inj_opt pat =
    match inj_opt with
    | None -> Walker_ast.Avar pat
    | Some inj_ty_ident -> Walker_ast.Ainj (get_inj_name inj_ty_ident, pat)

  and make_walker_expr inj_opt pat =

    let rec make_expr pat =
      match pat with
      | Walker_ast.Pvariable (_, type_expr) ->
        let walker = walker_fun type_expr in
        let arguments =
          match bufferize_pattern pat with
          | None, p ->
            [ argument_of_pat inj_opt p ]
          | Some b, p ->
            [ argument_of_pat None b; argument_of_pat inj_opt p ] in
        Walker_ast.Eapply (walker, arguments)
      | Walker_ast.Ptuple (pat, _type_expr) ->
        Walker_ast.Etuple (Array.map make_expr pat) in

    make_expr pat

  and make_walker_clause ty =
    let pat = make_pattern ty in
    Walker_ast.Clause (pat, make_walker_expr None pat)

  and make_walker_tuple_fun ty =
    let pat = make_pattern ty in
    let walker_expr = make_walker_expr None pat in
    let pats =
      match bufferize_pattern pat with
      | None, p -> [ p ]
      | Some b, p -> [ b; p; ] in
    Walker_ast.Fabs (pats, walker_expr)

  and make_walker_alias_clause ty_name vis alias_type_expr =

    let alias_pat = make_pattern alias_type_expr in
    let alias_expression = make_walker_expr None alias_pat in

    match vis with
    | Abstract | Public ->
      Walker_ast.Clause (alias_pat, alias_expression)

    | Private ->
      let ty_pat = make_pattern (Const_type (Lident ty_name)) in
      let ty_inj_ident = Lident ty_name in

      let arguments = [ argument_of_pat (Some ty_inj_ident) ty_pat ] in

      let ty_expression =
        Walker_ast.Eapply (
          Walker_ast.Fabs ([alias_pat], alias_expression),
          arguments) in

      Walker_ast.Clause (ty_pat, ty_expression)
  ;;

  let record_field = function
    | Field (label, ty)
    | Field_mutable (label, ty) ->
      Walker_ast.Record_field (label, make_walker_clause ty)
  ;;

  let sum_field = function
    | Sum_tag sum_label ->
      Walker_ast.Sum_tag sum_label
    | Sum_inj (sum_label, ty) ->
      Walker_ast.Sum_inj (sum_label, make_walker_clause ty)
  ;;

  let start_clause () = Symbol.reset ();;

  let rec type_kind ty_name = function
    | Record_type (vis, record_fields) ->
      start_clause ();
      Walker_ast.Record
        (buffer_argument, vis, List.map record_field record_fields)
    | Sum_type (vis, sum_fields) ->
      let sum_clause sf =
        start_clause ();
        sum_field sf in
      Walker_ast.Sum
        (buffer_argument, vis, List.map sum_clause sum_fields)
    | Alias_type (vis, alias_field) ->
      start_clause ();
      Walker_ast.Alias
        (buffer_argument, vis, alias_clause ty_name vis alias_field)
    | External_type ->
      let ty_abstract = Const_type (Lident "abstract") in
      type_kind ty_name (Alias_type (Abstract, (ty_abstract, None)))

  and alias_clause ty_name vis (type_expr, kind_opt as _alias_field) =
    let kind_opt =
      match kind_opt with
      | None -> None
      | Some ty_kind -> Some (type_kind ty_name ty_kind) in
    make_walker_alias_clause ty_name vis type_expr, kind_opt
  ;;

  let make_walker_var_arg_name ty_kind =

    let used_ty_arguments =
      find_ty_vars_in_ty_kind ty_kind in

    (fun ty_argument ->
     if List.mem ty_argument used_ty_arguments
     then walker_var_arg_name ty_argument
     else walker_unused_var_arg_name ty_argument)
  ;;

  let type_def ((ty_name, walker_name), ty_arguments, ty_kind) =
    let arg_names =
      List.map (make_walker_var_arg_name ty_kind) ty_arguments in
    ((ty_name, walker_name), arg_names, type_kind ty_name ty_kind)
  ;;

  let type_defs ty_def_list =
    let gen_walker_name (ty_name, ty_arguments, ty_kind) =
      let walker_name = local_walker_name ty_name in
      let binding = (ty_name, walker_name) in
      local_env := binding :: !local_env;
      binding, ty_arguments, ty_kind in
    let gen_walker_defs =
      List.map gen_walker_name ty_def_list in

    List.map type_def gen_walker_defs
  ;;

  let phrase accu = function
    | Type_definition ty_def_list ->
      let tdefs = type_defs ty_def_list in
      let rec_flag = rec_flag_of_type_definition ty_def_list in
      Walker_ast.Type_definition (rec_flag, tdefs) :: accu
    | Comment s -> Walker_ast.Comment s :: accu
    | Directive (Dir_open m) ->
      let d = Walker_ast.Dir_open (external_walker_module_name m) in
      Walker_ast.Directive d :: accu
    | Directive (Dir_include m) ->
      let d = Walker_ast.Dir_include (external_walker_module_name m) in
      Walker_ast.Directive d :: accu
  ;;

  let file phs =
    let res = List.rev (List.fold_left phrase [] phs) in
    if Og_debug.get_debug () then
      Format.fprintf
        Format.err_formatter "%a@." Walker_ast_print.print_file res;
    res
  ;;

end
;;
