(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 1997-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: lib_read.ml,v 1.11 2014/01/31 09:51:52 fclement Exp $ *)

(* Runtime library for readers generated with OCamlGen. *)

open Scanf;;

open Scanf_reader;;

(* Printing tokens that must be escaped.
   These escapes are not in the sense of String.escaped or
   Char.escaped, since accented chars should not be replaced by the
   decimal encoding of their ascii code. *)

let id x = x;;

let read_unit ib =
  Scanf.bscanf ib "%, ( )" ()
;;

(* Reading strings with quotes. *)
let read_quoted_string ib = Scanf.bscanf ib "%S" id;;

let read_quoted_char ib = bscanf ib "%C" id;;
let read_quoted_int ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %i )" id
  | _ -> Scanf.bscanf ib "%i" id)
;;
let read_quoted_nativeint ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %ni )" id
  | _ -> Scanf.bscanf ib "%ni" id)
;;
let read_quoted_int32 ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %li )" id
  | _ -> Scanf.bscanf ib "%li" id)
;;
let read_quoted_int64 ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %Li )" id
  | _ -> Scanf.bscanf ib "%Li" id)
;;
let read_quoted_float ib =
  Scanf.bscanf ib "%0c" (function
  | '(' -> (* ) *)
    Scanf.bscanf ib "( %f )" id
  | _ -> Scanf.bscanf ib "%f" id)
;;

(* Iterators *)
let read_List f ib =
  let rec rl accu ib =
    Scanf.bscanf ib "%0c" (function
    | (* [ *) ']' ->
      Scanf.bscanf ib "]" accu
    | _ ->
      Scanf.bscanf ib "%r ; " f
      (fun x -> rl (x :: accu) ib)) in
  bscanf ib "[ %r ]" (rl []) (fun l -> List.rev l)
;;

let read_Array f ib =
  let rec rl accu ib =
    Scanf.bscanf ib "%0c" (function
    | '|' ->
      Scanf.bscanf ib "|]" accu
    | _ ->
      Scanf.bscanf ib "%r ; " f
      (fun x -> rl (x :: accu) ib)) in
  bscanf ib "[| %r |]" (rl []) (fun l -> Array.of_list (List.rev l))
;;

let read_Option f ib =
  Scanf.bscanf ib "%0c" (function
  | 'N' ->
    Scanf.bscanf ib "None" None
  | _ ->
    Scanf.bscanf ib "Some %r" f (fun x -> Some x))
;;

let _read_list fmt (f : 'a reader) : 'a list reader = fun ib ->
  let rec rl accu ib =
    Scanf.kscanf ib (fun _ _ -> accu) "%r" f (fun x ->
      let accu = x :: accu in
      Scanf.kscanf ib (fun _ _ -> accu) fmt
        (rl accu) ib
    ) in
  List.rev (rl [] ib)
;;

let read_list fmt f ib =

  let rec loop accu ib =
    try
      let accu = f ib :: accu in
      try
        Scanf.bscanf ib fmt;
        loop accu ib
      with _ -> accu
    with _ -> accu in

  List.rev (loop [] ib)
;;

let read_array fmt f ib =
  let rec loop accu ib =
    try
      let accu = f ib :: accu in
      try
        Scanf.bscanf ib fmt;
        loop accu ib
      with _ -> accu
    with _ -> accu in

  Array.of_list (List.rev (loop [] ib))
;;

let read_option fmt (f : (Scanf.Scanning.in_channel -> 'a)) ib : 'a option =
  try
    Scanf.bscanf ib fmt;
    Some (f ib)
  with _ -> None
;;

let read_bool ib =
  Scanf.bscanf ib "%0c" (function
  | 't' ->
    Scanf.bscanf ib "true" true
  | _ ->
    Scanf.bscanf ib "false" false)
;;

let read_poly ib =
  bscanf ib "<poly>" (failwith "Cannot read a polymorphic value")
;;

let read_fun ib =
  bscanf ib "<fun>" (fun _ -> failwith "Cannot read a polymorphic function")
;;
